import { readdir, writeFile } from 'fs/promises'
import process from 'process'

interface AnnounceOpts {
  /** Output filename */
  out: string
  releaseDir: string
}

interface ReleasedApkg {
  name: string
  url: string
}

export async function announce (opts: AnnounceOpts): Promise<void> {
  const version = getReleaseVersion(opts.releaseDir)
  const announcement = [
    `# Sinologie Anki Pack ${version} ist released! 🎉`,
    `## Details zum Release:\n${linkToReleaseDetails(version)}`,
    `## Download komplettes Pack als ZIP:\n${
      linkToCombinedZip(opts.releaseDir)
    }`,
    '## Download einzelne APKGs:',
    ...(await linksToApkgs(opts.releaseDir))
  ].join('\n\n')
  await writeFile(opts.out, announcement, { encoding: 'utf-8' })
}

function getReleaseVersion (dir: string): string {
  const dashIdx = dir.lastIndexOf('-')
  if (dashIdx < 0) {
    throw new Error(`${dir} does not look like a release dir`)
  }
  return dir.substring(dashIdx + 1)
}

function linkToReleaseDetails (version: string): string {
  return (
    `https://gitlab.phaidra.org/kartenaale/sinologie-anki-pack/-/releases/${
      version
    }`
  )
}

function linkToCombinedZip (releaseDir: string): string {
  return rawArtifactDownload(`${releaseDir}.zip`)
}

async function linksToApkgs (releaseDir: string): Promise<string[]> {
  return (await findReleasedApkgs(releaseDir))
    .map(({ name, url }) => {
      return `### ${name}\n${url}`
    })
}

async function findReleasedApkgs (releaseDir: string): Promise<ReleasedApkg[]> {
  return (await readdir(releaseDir))
    .filter(e => e.endsWith('.apkg'))
    .map(path => {
      return {
        name: path.substring(0, path.length - '.apkg'.length),
        url: rawArtifactDownload(path)
      }
    })
}

function rawArtifactDownload (path: string): string {
  if (path === '') {
    throw new Error('path required')
  }
  return `https://gitlab.phaidra.org/kartenaale/sinologie-anki-pack/-/jobs/${
    getCiJobId()
  }/artifacts/raw/${encodeURIComponent(path)}`
}

function getCiJobId (): string {
  return process.env.CI_JOB_ID ?? 'job-id-unavailable'
}
