import cnchar from 'cnchar'
import cncharOrder from 'cnchar-order'
import cncharRadical from 'cnchar-radical'
import cncharTrad from 'cnchar-trad'
import hanzi from 'hanzi-tool'
import { loadHanziWriterData } from './hanzi-writer'

export interface TraditionalInfo {
  /**
   * Radical as full character, e.g. distinguishing mound and city.
   *
   * `?` if no data available.
   */
  radical: string
  /**
   * Stroke count of the character, by traditional counting rules,
   * i.e. using the full character form of the radical.
   *
   * `?` if no data available.
   */
  count: string
}

/** If cnchar does not know the char */
interface CharOverride {
  extraStrokes: number
  radical: string
}
const charOverrides = new Map<string, CharOverride>()
charOverrides.set('璮', { radical: '玉', extraStrokes: 13 })
charOverrides.set('矊', { radical: '目', extraStrokes: 14 })
charOverrides.set('虝', { radical: '虍', extraStrokes: 6 })
charOverrides.set('複', { radical: '衣', extraStrokes: 9 })
// cnchar thinks this has ba (8)
charOverrides.set('義', { radical: '羊', extraStrokes: 7 })
// no idea why 13 is correct here
// https://www.mdbg.net/chinese/dictionary?page=worddict&wdrst=1&wdqb=%E9%9A%A8
charOverrides.set('隨', { radical: '阜', extraStrokes: 13 })
// no idea why 13 is correct here
// https://www.mdbg.net/chinese/dictionary?page=worddict&wdrst=1&wdqb=%E9%A4%90
charOverrides.set('餐', { radical: '食', extraStrokes: 7 })
charOverrides.set('了', { radical: '亅', extraStrokes: 1 })

export async function getTraditionalInfo (
  char: string
): Promise<TraditionalInfo> {
  if (char.length !== 1) {
    throw new Error('single chars only')
  }
  startOnce()

  const override = charOverrides.get(char)
  if (override !== undefined) {
    const radicalCount = await getRadicalStrokesFull(override.radical)
    return {
      radical: override.radical,
      count: `${radicalCount}+${override.extraStrokes}`
    }
  }

  const radicalInline = getRadicalInline(char)
  const radical = getRadicalFullForm(char, radicalInline)
  const radicalStrokesInline = await getRadicalStrokesInline(radicalInline)
  const radicalStrokesFull = await getRadicalStrokesFull(radical)
  const totalStrokes = await getTotalStrokes(char)
  const extra = Math.max(0, totalStrokes - radicalStrokesInline)
  return {
    radical,
    count: `${radicalStrokesFull}+${extra}`
  }
}

let started = false
function startOnce (): void {
  if (!started) {
    cnchar.use(cncharOrder, cncharRadical, cncharTrad)
    hanzi.start()
    started = true
  }
}

function getRadicalInline (char: string): string {
  const radical = cnchar.radical(char)[0].radical
  if (radical === '') {
    console.error(`Cannot get radical for: ${char}`)
  }
  return radical
}

/**
 * If cnchar reports a different radical but there is an easy mapping to
 * the correct one.
 *
 * Also normalizes variant forms, see:
 * https://www.xiaoma.info/bushou.php?ext=1
 */
const radicalToFull = new Map()
radicalToFull.set('艹', '艸')
radicalToFull.set('王', '玉')
radicalToFull.set('⺩', '玉')
radicalToFull.set('忄', '心')
radicalToFull.set('㣺', '心')
radicalToFull.set('衤', '衣')
radicalToFull.set('⺮', '竹')
radicalToFull.set('糹', '糸')
radicalToFull.set('纟', '糸')
radicalToFull.set('攵', '攴')
radicalToFull.set('刂', '刀')
radicalToFull.set('丷', '八')
radicalToFull.set('亻', '人')
radicalToFull.set('乀', '丿')
radicalToFull.set('乁', '丿')
radicalToFull.set('乚', '乙')
radicalToFull.set('乛', '乙')
radicalToFull.set('㔾', '卩')
radicalToFull.set('尣', '尢')
radicalToFull.set('巛', '川')
radicalToFull.set('巜', '川')
radicalToFull.set('彑', '彐')
radicalToFull.set('户', '戶')
radicalToFull.set('戸', '戶')
radicalToFull.set('扌', '手')
radicalToFull.set('龵', '手')
radicalToFull.set('歺', '歹')
radicalToFull.set('母', '毋')
radicalToFull.set('⺟', '毋')
radicalToFull.set('氵', '水')
radicalToFull.set('氺', '水')
radicalToFull.set('灬', '火')
radicalToFull.set('爫', '爪')
radicalToFull.set('丬', '爿')
radicalToFull.set('牜', '牛')
radicalToFull.set('甩', '用')
radicalToFull.set('⺪', '疋')
radicalToFull.set('礻', '示')
radicalToFull.set('罓', '网')
radicalToFull.set('罒', '网')
radicalToFull.set('⺳', '网')
radicalToFull.set('⺶', '羊')
radicalToFull.set('耂', '老')
radicalToFull.set('肀', '聿')
radicalToFull.set('月', '肉') // REVIEW: exceptions?
radicalToFull.set('襾', '西')
radicalToFull.set('覀', '西')
radicalToFull.set('见', '見')
radicalToFull.set('訁', '言')
radicalToFull.set('讠', '言')
radicalToFull.set('贝', '貝')
radicalToFull.set('赱', '走')
radicalToFull.set('⻊', '足')
radicalToFull.set('车', '車')
radicalToFull.set('辶', '辵')
radicalToFull.set('釒', '金')
radicalToFull.set('钅', '金')
radicalToFull.set('长', '長')
radicalToFull.set('门', '門')
radicalToFull.set('靑', '青')
radicalToFull.set('靣', '面')
radicalToFull.set('韦', '韋')
radicalToFull.set('页', '頁')
radicalToFull.set('风', '風')
radicalToFull.set('飞', '飛')
radicalToFull.set('飠', '食')
radicalToFull.set('饣', '食')
radicalToFull.set('马', '馬')
radicalToFull.set('髙', '高')
radicalToFull.set('鱼', '魚')
radicalToFull.set('鸟', '鳥')
radicalToFull.set('鹵', '卤')
radicalToFull.set('麦', '麥')
radicalToFull.set('黄', '黃')
radicalToFull.set('黾', '黽')
radicalToFull.set('鼔', '鼓')
radicalToFull.set('齐', '齊')
radicalToFull.set('齿', '齒')
radicalToFull.set('龙', '龍')
radicalToFull.set('龟', '龜')
function getRadicalFullForm (char: string, radicalInline: string): string {
  if (radicalInline === '阝') {
    // special case: if on the right, it means 邑 (city),
    //               but on the left it means 阜 (mound)
    // we check if the character starts with 𠄎 or ㇌ and if so, assume it means
    // mound
    const firstStrokeShape =
        cnchar.stroke(char, 'order', 'shape')[0][0] as string
    if (firstStrokeShape.startsWith('𠄎') || firstStrokeShape.endsWith('㇌')) {
      // probably on the left
      return '阜'
    } else {
      // probably on the right
      return '邑'
    }
  }
  return radicalToFull.get(radicalInline) ?? radicalInline
}

async function getRadicalStrokesFull (radical: string): Promise<number> {
  const hanziWriterStrokeCount = await getStrokesFromHanziWriterData(radical)
  if (hanziWriterStrokeCount !== undefined) {
    // use stroke count from animation if available
    return hanziWriterStrokeCount
  }

  if ((cnchar.radical as any).isRadical(radical) === true) {
    return (cnchar.radical as any).getRadicalCount(radical)
  }
  // radical as full char differs from inline form, count strokes
  const count = cnchar.stroke(radical, 'array')[0] as number
  if (count === 0) {
    console.error(`zero stroke count for full form of radical: ${radical}`)
  }
  return count
}

async function getRadicalStrokesInline (
  radicalInline: string
): Promise<number> {
  if (radicalInline === '艹') {
    // in Taiwan this is written with 4 strokes
    return 4
  }
  if (radicalInline === '糹') {
    // not defined in cnchar
    return 6
  }
  const hanziWriterStrokeCount =
      await getStrokesFromHanziWriterData(radicalInline)
  if (hanziWriterStrokeCount !== undefined) {
    // use stroke count from animation if available
    return hanziWriterStrokeCount
  }
  let count = (cnchar.radical as any).getRadicalCount(radicalInline) as number
  if (count === 0) {
    count = cnchar.stroke(radicalInline, 'array')[0] as number
  }
  if (count === 0) {
    console.error(
      `zero stroke count for inline form of radical: ${radicalInline}`
    )
  }
  return count
}

/**
 * Gets the total stroke count from cnchar and compensates for Taiwan variants,
 * e.g. one stroke is added for each grass radical which is written with four
 * strokes instead of three in Taiwan.
 */
async function getTotalStrokes (char: string): Promise<number> {
  const hanziWriterStrokeCount = await getStrokesFromHanziWriterData(char)
  if (hanziWriterStrokeCount !== undefined) {
    // use stroke count from animation if available
    return hanziWriterStrokeCount
  }

  let count = cnchar.stroke(char, 'array')[0] as number
  if (count === 0) {
    const fallback = await getStrokesFromHanziWriterData(char)
    if (fallback !== undefined) {
      return fallback
    }
    console.error(`zero stroke count for character: ${char}`)
  }

  const parts = hanzi.decompose(char, 2).components as string[]
  for (const part of parts) {
    if (part === '艹') {
      count += 1
    }
  }
  return count
}

async function getStrokesFromHanziWriterData (
  char: string
): Promise<number | undefined> {
  const data = await loadHanziWriterData(char)
  if (data === undefined) {
    return undefined
  }
  return data.strokes.length
}
